import argparse
from pathlib import Path
import os
from .handlers.request_folder import request_folder
from .utils import is_empty, copy_from_installed_wheel
from .library import list_component_library, install_library, fetch_library, save_component_library_config
from .compiler import compile
import json

from .logger.builder import get_logger

LOGGER = get_logger(__name__)

def init_xircuits():
    package_name = 'xircuits'
    copy_from_installed_wheel(package_name, resource='.xircuits', dest_path='.xircuits')
    component_library_path = Path(os.getcwd()) / "xai_components"
    if not component_library_path.exists():
        copy_from_installed_wheel('xai_components', '', 'xai_components')

def cmd_start_xircuits(args, extra_args=[]):
    # fetch xai_components
    component_library_path = Path(os.getcwd()) / "xai_components"
    if not component_library_path.exists():
        copy_from_installed_wheel('xai_components', '', 'xai_components')

    news_url_option = '--LabApp.news_url="https://xpress.ai/blog/atom.xml"'

    # handler for extra jupyterlab launch options
    if extra_args:
        try:
            launch_cmd = "jupyter lab --ContentsManager.allow_hidden=True" + " " + " ".join(extra_args) + " " + news_url_option
            os.system(launch_cmd)
        except Exception as e:
            LOGGER.error("Error in launch args! Error log:\n")
            LOGGER.error(e)
    else:
        os.system(f"jupyter lab --ContentsManager.allow_hidden=True {news_url_option}")

def cmd_download_examples(args, extra_args=[]):
    if not os.path.exists("examples") or is_empty("examples"):
        copy_from_installed_wheel('examples')
        LOGGER.info("Example workflows ready at working directory.")

def cmd_fetch_library(args, extra_args=[]):
    fetch_library(args.library_name)

def cmd_install_library(args, extra_args=[]):
    install_library(args.library_name)

def cmd_compile(args, extra_args=[]):
    component_paths = {}
    if args.python_paths_file:
        component_paths = json.load(args.python_paths_file)
    compile(args.source_file, args.out_file, component_python_paths=component_paths)

def cmd_list_libraries(args, extra_args=[]):
    list_component_library()

def main():
    parser = argparse.ArgumentParser(description='Xircuits Command Line Interface', add_help=False)
    subparsers = parser.add_subparsers(dest="command")

# Adding parser for 'start' command
    start_parser = subparsers.add_parser('start', help='Start Xircuits.')
    # Add an arbitrary list of arguments. The nargs="*" means 0 or more arguments.
    # This will collect all additional arguments into a list.
    start_parser.add_argument('extra_args', nargs='*', help='Additional arguments for Xircuits launch command')
    start_parser.set_defaults(func=cmd_start_xircuits)

    # Adding parser for 'install' command
    install_parser = subparsers.add_parser('install', help='Fetch and installs a library for Xircuits.')
    install_parser.add_argument('library_name', type=str, help='Name of the library to install')
    install_parser.set_defaults(func=cmd_install_library)

    # Adding parser for 'fetch' command
    fetch_parser = subparsers.add_parser('fetch-only', help='Fetch a library for Xircuits. Does not install.')
    fetch_parser.add_argument('library_name', type=str, help='Name of the library to fetch')
    fetch_parser.set_defaults(func=cmd_fetch_library)

    # Adding parser for 'examples' command
    examples_parser = subparsers.add_parser('examples', help='Get example workflows for Xircuits.')
    examples_parser.add_argument('--branch', nargs='?', default="master", help='Load example workflows to current working directory/')
    examples_parser.set_defaults(func=cmd_download_examples)

    # Adding parser for 'compile' command
    compile_parser = subparsers.add_parser('compile', help='Compile a Xircuits workflow file.')
    compile_parser.add_argument('source_file', type=argparse.FileType('r', encoding='utf-8'))
    compile_parser.add_argument('out_file', type=argparse.FileType('w', encoding='utf-8'))
    compile_parser.add_argument("python_paths_file", nargs='?', default=None, type=argparse.FileType('r'),
                                help="JSON file with a mapping of component name to required python path. "
                                     "e.g. {'MyComponent': '/some/path'}")
    compile_parser.set_defaults(func=cmd_compile)

    # Adding parser for 'list' command
    list_parser = subparsers.add_parser('list', help='List available component libraries for Xircuits.')
    list_parser.set_defaults(func=cmd_list_libraries)


    args, unknown_args = parser.parse_known_args()

    if hasattr(args, 'func'):
        args.func(args, unknown_args)
    else:
        valid_help_args = {"-h", "--h" "-help", "--help"}
        if any(arg in unknown_args for arg in valid_help_args):
            parser.print_help()
        else:
            # Default behavior: if no sub-command is provided, start xircuits.
            cmd_start_xircuits(args, unknown_args)

    return 0

if __name__ == '__main__':
    main()

LOGGER.info(
'''
======================================
__   __  ___                _ _       
\ \  \ \/ (_)_ __ ___ _   _(_) |_ ___ 
 \ \  \  /| | '__/ __| | | | | __/ __|
 / /  /  \| | | | (__| |_| | | |_\__ \\
/_/  /_/\_\_|_|  \___|\__,_|_|\__|___/
                                      
======================================
'''
)

config_path = Path(os.getcwd()) / ".xircuits"
if not config_path.exists():
    init_xircuits()

save_component_library_config()


from xai_components.base import InArg, OutArg, InCompArg, Component, BaseComponent, xai_component, dynalist, SubGraphExecutor

@xai_component(type='branch')
class BranchComponent(Component):
    """A component that conditionally executes one of two branches based on a boolean condition.
    
    ##### inPorts:
    - condition (bool): A boolean condition to evaluate.
    
    ##### Branches:
    - when_true: Branch that executes if the condition is True.
    - when_false: Branch that executes if the condition is False.
    """
    when_true: BaseComponent
    when_false: BaseComponent

    condition: InArg[bool]
    
    def do(self, ctx) -> BaseComponent:
        next = None
        if self.condition.value:
            if hasattr(self, 'when_true'):
                next = self.when_true
        else:
            if hasattr(self, 'when_false'):
                next = self.when_false
        while next:
            next = next.do(ctx)
        if hasattr(self, 'next') and self.next:
            return self.next
    
@xai_component(type='branch')
class LoopComponent(Component):
    """A component that loops while the condition is True.
    
    ##### inPorts:
    - condition (bool): Boolean that determines whether `body` branch is executed.
    
    ##### Branches:
    - body: Branch that executes in each iteration of the loop.
    """
    body: BaseComponent
    condition: InArg[bool]
    
    def do(self, ctx) -> BaseComponent:
        while self.condition.value:
            next_body = self.body.do(ctx)
            while next_body:
                next_body = next_body.do(ctx)
            return self
        if hasattr(self, 'next') and self.next:
            return self.next

@xai_component(type='branch')
class ReverseForEach(Component):
    """A component that iterates over a list in reverse order, executing body branches for each item.
    
    ##### inPorts:
    - items (list): The list of items to iterate over.
    
    ##### outPorts:
    - current_item (any): The current item in the iteration.
    - current_index (int): The index of the current item in the iteration.
    
    ##### Branches:
    - body: Branch that executes for each item.
    """
    body: BaseComponent
    items: InCompArg[list]
    current_item: OutArg[any]
    current_index: OutArg[int]
    
    def do(self, ctx) -> BaseComponent:
        for i, item in enumerate(self.items.value[-1]):
            self.current_item.value = item
            self.current_index.value = i
            
            next_body = self.body.do(ctx)
            while next_body:
                next_body = next_body.do(ctx)
        if hasattr(self, 'next') and self.next:
            return self.next

@xai_component(type='branch')
class ForEach(Component):
    """A component that iterates over a list, executing body branches for each item.
    
    ##### inPorts:
    - items (list): The list of items to iterate over.
    
    ##### outPorts:
    - current_item (any): The current item in the iteration.
    - current_index (int): The index of the current item in the iteration.
    
    ##### Branches:
    - body: Branch that executes for each item.
    """
    body: BaseComponent
    items: InCompArg[list]
    current_item: OutArg[any]
    current_index: OutArg[int]
    
    def do(self, ctx) -> BaseComponent:
        for i, item in enumerate(self.items.value):
            self.current_item.value = item
            self.current_index.value = i
            
            next_body = self.body.do(ctx)
            while next_body:
                next_body = next_body.do(ctx)
        if hasattr(self, 'next') and self.next:
            return self.next

@xai_component
class CounterComponent(Component):
    """A component that maintains and increments a counter.
    
    ##### inPorts:
    - start_number (int): The initial value of the counter.
    - step (int): The increment step for the counter.
    
    ##### outPorts:
    - out_number (int): The current value of the counter.
    """
    start_number: InArg[int]
    step: InArg[int]
    out_number: OutArg[int]
    state: any
    
    def __init__(self):
        super().__init__()
        self.state = None
        
    def execute(self, ctx) -> None:
        if self.state is None:
            self.state = self.start_number.value
        else:
            self.state += self.step.value
            
        self.out_number.value = self.state

@xai_component
class ComparisonComponent(Component):
    """A component that performs a comparison between two integers.
    
    ##### inPorts:
    - a (int): The first integer.
    - b (int): The second integer.
    - op (str): The comparison operator (as a string).
    
    ##### outPorts:
    - out (bool): The result of the comparison (boolean).
    """
    a: InArg[int]
    b: InArg[int]
    op: InArg[str]
    out: OutArg[bool]

    def execute(self, ctx) -> None:
        self.out.value = eval(str(self.a.value) + " " + self.op.value + " " + str(self.b.value))


class MutableVariable:
    _fn: any

    def __init__(self):
        self._fn = None
    
    def set_fn(self, fn) -> None:
        self._fn = fn
        
    @property
    def value(self) -> any:
        return self._fn()
    
    
@xai_component(type='context_get')
class GetVariableComponent(Component):
    """A component that retrieves a variable from the context.
    
    ##### inPorts:
    - name (str): The name of the variable to retrieve.
    
    ##### outPorts:
    - value (any): The value of the retrieved variable.
    """
    name: InArg[str]
    value: OutArg[any]
    
    def __init__(self):
        super().__init__()
        self.value = MutableVariable()
        
    def execute(self, ctx) -> None:
        self.value.set_fn(lambda: ctx[self.name.value])


@xai_component(type='context_set')
class SetVariableComponent(Component):
    """A component that sets a variable in the context.
    
    ##### inPorts:
    - name (str): The name of the variable to set.
    - value (any): The value to set for the variable.
    """
    name: InArg[str]
    value: InArg[any]

    def execute(self, ctx) -> None:
        ctx[self.name.value] = self.value.value


@xai_component(type='variable')
class DefineVariableComponent(Component):
    """A component that defines a variable in the context and creates a reference to it.
    
    ##### inPorts:
    - name (str): The name of the variable to define.
    - value (any): The initial value of the variable.
    
    ##### outPorts:
    - ref (any): A reference to the defined variable.
    """
    name: InArg[str]
    value: InArg[any]
    ref: OutArg[any]

    def __init__(self):
        super().__init__()
        self.ref = MutableVariable()
        
    def execute(self, ctx) -> None:
        ctx[self.name.value] = self.value.value
        self.ref.set_fn(lambda: ctx[self.name.value])


@xai_component
class EvalBooleanExpression(Component):
    """A component that evaluates a boolean expression.
    
    ##### inPorts:
    - expression (str): The boolean expression to evaluate.
    - args (dynalist): A list of arguments to use in the expression.
    
    ##### outPorts:
    - out (bool): The result of the evaluated expression (boolean).
    """
    expression: InCompArg[str]
    args: InArg[dynalist]
    out: OutArg[bool]

    def execute(self, ctx) -> None:
        args = []
        for arg in self.args.value:
            if hasattr(arg, 'value'):
                args.append(arg.value)
            else:
                args.append(arg)

        exec('self.out.value = ( ' + self.expression.value + ')', globals(), locals())

@xai_component
class EvaluateExpressionWithVariables(Component):
    """Evaluates a single mathematical or boolean expression with variables from a dictionary and context.

    ##### inPorts:
    - expression (str): A mathematical or boolean expression as a string with placeholders for variables.
    - values_dict (dict): A dictionary containing variable values.

    ##### outPorts:
    - result (float): The result after evaluating the expression with the provided values.

    ##### Description:
    This component evaluates an expression using variables from both the provided dictionary and the context (`ctx`). The dictionary values take priority over the context values if there are any conflicts.

    ##### Example:
      - expression: "{a} + {b} * {c}"
      - ctx: {"a": 1, "b": 2}
      - values_dict: {"c": 3}
    """
    expression: InArg[str]
    values_dict: InArg[dict]
    result: OutArg[float]

    def execute(self, ctx) -> None:
        expression = self.expression.value
        values_dict = self.values_dict.value if self.values_dict.value else {}

        combined_dict = {**ctx, **values_dict}

        formatted_expression = expression.format(**combined_dict)
        print(f'Expression : {formatted_expression}')

        result = eval(formatted_expression)
        self.result.value = result
        print("Evaluation Result:", result)

@xai_component(color='blue')
class ExceptionHandler(Component):
    """Executes a body branch and handles any exceptions that occur.

    ##### Branches:
    - body: The main execution path of components to execute.
    - handler: The execution path if an exception occurs.

    ##### outPorts:
    - exception (str): The exception message if an exception is caught.
    """
    body: BaseComponent
    handler: BaseComponent
    exception: OutArg[str]

    def execute(self, ctx):
        try:
            SubGraphExecutor.do(self.body, ctx)
        except Exception as e:
            self.exception.value = str(e)
            print(e)
            SubGraphExecutor.do(self.handler, ctx)


# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2024, TVB Widgets Team
#

from tvb.datatypes.connectivity import Connectivity
from xai_components.base import InArg, OutArg, xai_component
from xai_components.base_tvb import TVBComponent
from xai_components.utils import print_component_summary


@xai_component(color='rgb(85, 37, 130)')
class ConnectivityFromFile(TVBComponent):
    file_path: InArg[str]

    connectivity: OutArg[Connectivity]

    def __init__(self):
        self.done = False
        self.file_path = InArg(None)
        self.connectivity = OutArg(None)

    def execute(self, ctx) -> None:
        # imports
        from matplotlib import pyplot as plt
        from tvb.datatypes.connectivity import Connectivity

        file_path = self.file_path.value
        if not file_path:
            file_path = 'connectivity_76.zip'  # default from tvb_data
        connectivity = Connectivity.from_file(file_path)
        connectivity.configure()

        self.connectivity.value = connectivity
        print_component_summary(self.connectivity.value)

        plt.imshow(self.connectivity.value.weights, interpolation='none')
        plt.show()



STORAGE_CONFIG_FILE = 'storage_config.json'  # To be used only for HPC runs
COLLAB_NAME_KEY = 'collab_name'  # Used only for HPC runs
BUCKET_NAME_KEY = 'bucket_name'  # Used only for HPC runs
FOLDER_PATH_KEY = 'folder_path'  # Used only for HPC runs
STORE_RESULTS_DIR = 'results'  # Used by component that takes care of storing data and stage-out from HPC
DIR_TIME_STAMP_FRMT = '%Y.%m.%d_%H_%M_%S'

import os
import urllib.parse
import shutil
import importlib_resources

def is_empty(directory):
    # will return true for uninitialized submodules
    return not os.path.exists(directory) or not os.listdir(directory)

def is_valid_url(url):
    try:
        result = urllib.parse.urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False

def copy_from_installed_wheel(package_name, resource="", dest_path=None):
    if dest_path is None:
        dest_path = package_name

    # Get the resource reference
    ref = importlib_resources.files(package_name) / resource

    # Create the temporary file context
    with importlib_resources.as_file(ref) as resource_path:
        shutil.copytree(resource_path, dest_path)


import subprocess
import sys
import os
import json
from pathlib import Path
from ..utils import is_valid_url, is_empty
from ..handlers.request_submodule import request_submodule_library
from ..handlers.request_folder import clone_from_github_url

def build_component_library_path(component_library_query: str) -> str:
    if "xai" not in component_library_query:
        component_library_query = "xai_" + component_library_query

    if "xai_components" not in component_library_query:
        component_library_query = "xai_components/" + component_library_query

    return component_library_query

def get_library_config(library_name, config_key):
    config_path = ".xircuits/component_library_config.json"
    if os.path.exists(config_path):
        with open(config_path, "r") as config_file:
            config = json.load(config_file)
            for library in config.get("libraries", []):
                if library.get("library_id") == library_name:
                    # Check for the existence of the key and that its value is not None
                    if config_key in library and library[config_key] is not None:
                        return library[config_key]
                    else:
                        return None  # Explicitly return None if the key is missing or its value is None
    
    return None  # Return None if the library isn't found or the file doesn't exist

def build_library_file_path_from_config(library_name, config_key):
    file_path = get_library_config(library_name, config_key)
    if file_path is None:
        return None

    base_path = get_library_config(library_name, "local_path")
    if base_path is None:
        return None
    
    full_path = Path(base_path, file_path)
    return full_path.as_posix()

def get_component_library_path(library_name: str) -> str:
    if is_valid_url(library_name):
        return clone_from_github_url(library_name)
    else:
        return build_component_library_path(library_name)

def install_library(library_name: str):
    print(f"Installing {library_name}...")
    component_library_path = get_component_library_path(library_name)

    if not Path(component_library_path).is_dir() or is_empty(component_library_path):
        success, message = request_submodule_library(component_library_path)
        if not success:
            print(message)
            return

    # Get the requirements path from the configuration
    requirements_path = get_library_config(library_name, "requirements_path")
    # Convert to Path object and ensure it's absolute
    requirements_path = Path(requirements_path).resolve() if requirements_path else Path(component_library_path) / "requirements.txt"

    # Install requirements if the file exists
    if requirements_path.exists():
        try:
            print(f"Installing requirements for {library_name}...")
            subprocess.run([sys.executable, "-m", "pip", "install", "-r", str(requirements_path)], check=True)
            print(f"Library {library_name} ready to use.")
        except Exception as e:
            print(f"An error occurred while installing requirements for {library_name}: {e}")
    else:
        print(f"No requirements.txt found for {library_name}. Skipping installation of dependencies.")
        print(f"Library {library_name} ready to use.")

def fetch_library(library_name: str):
    print(f"Fetching {library_name}...")
    component_library_path = get_component_library_path(library_name)

    if not Path(component_library_path).is_dir() or is_empty(component_library_path):
        success, message = request_submodule_library(component_library_path)
        if success:
            print(f"{library_name} library fetched and stored in {component_library_path}.")
        else:
            print(message)
    else:
        print(f"{library_name} library already exists in {component_library_path}.")


from xai_components.base import SubGraphExecutor, InArg, OutArg, Component, xai_component

@xai_component(type="Start", color="red")
class OnEvent(Component):
    """A component that listens for a specific event and triggers when the event occurs.
    
    ##### inPorts:
    - eventName (str): The name of the event to listen for.
    
    ##### outPorts:
    - payload (dict): The payload of the event when it is triggered.
    """
    eventName: InArg[str]
    payload: OutArg[dict]

    def init(self, ctx):
        ctx.setdefault('events', {}).setdefault(self.eventName.value, []).append(self)

@xai_component
class FireEvent(Component):
    """A component that fires an event, triggering any components listening for that event.
    
    ##### inPorts:
    - eventName (str): The name of the event to fire.
    - payload (dict): The payload of the event to fire.
    """
    eventName: InArg[str]
    payload: InArg[dict]

    def execute(self, ctx):
        eventListeners = ctx['events'][self.eventName.value]
        for listener in eventListeners:
            listener.payload.value = self.payload.value
            SubGraphExecutor(listener).do(ctx)



import os.path
import shutil
import pytest
from tvbextxircuits import nb_generator
from tvbextxircuits.nb_generator import NotebookGenerator, NotebookFactory
from xai_components.xai_tvb_simulator.simulator import Simulator


TEMP_DIR = "temp_test_dir"
nb_generator.NOTEBOOKS_DIR = TEMP_DIR


def test_notebook_generator():
    notebook_generator = NotebookGenerator(Simulator, '', {})
    assert os.path.exists(nb_generator.NOTEBOOKS_DIR)
    assert notebook_generator.notebook is not None
    assert len(notebook_generator.notebook['cells']) == 0

    notebook_generator.add_code_cell("print('Hello world')")
    assert len(notebook_generator.notebook['cells']) == 1

    notebook_generator.add_markdown_cell("Hello world")
    assert len(notebook_generator.notebook['cells']) == 2

    with pytest.raises(NotImplementedError):
        notebook_generator.get_notebook()


def test_widget_code_generator():
    notebook_factory = NotebookFactory()
    notebook = notebook_factory.get_notebook_for_component('Simulator', 'sim_id',
                                                           'xai_components/xai_tvb_simulator/simulator.py', {},
                                                           'xircuits_id')
    assert notebook is None

    notebook_factory = NotebookFactory()
    notebook = notebook_factory.get_notebook_for_component('ReducedWongWang', 'model_id',
                                                           'xai_components/xai_tvb_models/wong_wang.py', {},
                                                           'xircuits_id')
    assert notebook is not None
    assert "models.ReducedWongWang(**{})" in notebook.cells[2]['source']
    assert "w.export_filename = r'model_model_id'" in notebook.cells[2]['source']

    assert len(os.listdir(nb_generator.NOTEBOOKS_DIR)) == 0
    notebook_path = notebook_factory.store(notebook, 'model', 'xircuits_id')
    assert os.path.exists(notebook_path)

    notebook_factory = NotebookFactory()
    notebook = notebook_factory.get_notebook_for_component('StoreResultsToDrive', 'store_id',
                                                           'xai_components/xai_storage/store_results.py', {},
                                                           'xircuits_id')
    assert notebook is not None
    assert 'TimeSeriesBrowser()' in notebook.cells[2]['source']

    notebook_factory = NotebookFactory()
    notebook = notebook_factory.get_notebook_for_component('StoreResultsToDrive', 'store_id',
                                                           'xai_components/xai_storage/store_results.py',
                                                           {'collab_name': 'collab', 'folder_path': 'folder'},
                                                           'xircuits_id')
    assert notebook is not None
    assert "TimeSeriesBrowser('collab', 'folder')" in notebook.cells[2]['source']


def teardown_function():
    if os.path.exists(nb_generator.NOTEBOOKS_DIR):
        shutil.rmtree(nb_generator.NOTEBOOKS_DIR)



from hatchling.builders.hooks.plugin.interface import BuildHookInterface
import shutil

class XircuitsBuildHook(BuildHookInterface):
    PLUGIN_NAME = 'tvbextxircuits'

    def initialize(self, version, build_data):
        shutil.copyfile('.gitmodules', 'tvbextxircuits/.xircuits/.gitmodules')



from pathlib import Path

def create_or_update_library(library_name: str, filename: str, content: str):
    library_path = Path(f"xai_components/xai_{library_name}")
    if library_path.exists():
        return save_component_to_library(content, library_path, filename)
    else:
        library_path.mkdir(parents=True, exist_ok=False)
        init_library_directory(library_path)
        return save_component_to_library(content, library_path, filename)
    
def init_library_directory(library_path: Path):
    # Create __init__.py file
    init_file_path = library_path / "__init__.py"
    with open(init_file_path, "w") as f:
        f.write("# This file is required to make Python treat the directories as containing packages")

    # Create README.md file
    readme_file_path = library_path / "README.md"
    with open(readme_file_path, "w") as f:
        f.write(f"# {library_path.name}\n\nThis is a custom library for Xircuits.")

def save_component_to_library(content: str, library_path: Path, filename: str = "new_component.py"):
    base_filename, file_extension = filename.rsplit('.', 1)
    component_file_path = library_path / filename
    counter = 1

    # Check if the file exists and generate a new filename by appending a counter
    while component_file_path.exists():
        component_file_path = library_path / f"{base_filename}_{counter}.{file_extension}"
        counter += 1

    with open(component_file_path, "w") as file:
        file.write(content)

    return f"Component saved as {component_file_path.name} in {library_path}"


# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2024, TVB Widgets Team
#
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.region_mapping import RegionMapping
from tvb.datatypes.surfaces import Surface

from xai_components.base import InArg, OutArg, xai_component, InCompArg
from xai_components.base_tvb import TVBComponent
from xai_components.utils import set_defaults, set_values, print_component_summary


@xai_component(color='rgb(120, 70, 31)')
class RegionMapping(TVBComponent):
    file_path: InArg[str]
    connectivity: InCompArg[Connectivity]
    surface: InCompArg[Surface]

    region_mapping: OutArg[RegionMapping]

    @property
    def tvb_ht_class(self):
        from tvb.datatypes.region_mapping import RegionMapping
        return RegionMapping

    def __init__(self):
        self.done = False
        self.file_path = InArg(None)
        set_defaults(self, self.tvb_ht_class)
        self.region_mapping = OutArg(None)

    def execute(self, ctx) -> None:
        file_path = self.file_path.value
        if not file_path:
            file_path = 'regionMapping_16k_76.txt'  # default from tvb_data
        region_mapping = self.tvb_ht_class.from_file(source_file=file_path)
        set_values(self, region_mapping)

        self.region_mapping.value = region_mapping
        print_component_summary(self.region_mapping.value)



import os
import json
import traceback
import tornado
import posixpath
from http import HTTPStatus
from jupyter_server.base.handlers import APIHandler
from tvbextxircuits.library import install_library, fetch_library, build_library_file_path_from_config, save_component_library_config, get_component_library_config, create_or_update_library

class InstallLibraryRouteHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        library_name = input_data.get("libraryName")

        if not library_name:
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.finish(json.dumps({"error": "Library name is required"}))
            return

        try:
            message = install_library(library_name)
            self.finish(json.dumps({"status": "OK", "message": message}))
        except RuntimeError as e:
            self.set_status(HTTPStatus.BAD_REQUEST)
            message = str(e)
            print(message)
            self.finish(json.dumps({"error": message}))
        except Exception as e:
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
            message = f"An unexpected error occurred: {traceback.format_exc()}"
            print(message)
            self.finish(json.dumps({"error": message}))

class FetchLibraryRouteHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        library_name = input_data.get("libraryName")

        if not library_name:
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.finish(json.dumps({"error": "Library name is required"}))
            return

        try:
            message = fetch_library(library_name.lower())
            self.finish(json.dumps({"status": "OK", "message": message}))
        except RuntimeError as e:
            self.set_status(HTTPStatus.BAD_REQUEST)
            message = str(e)
            self.finish(json.dumps({"error": message}))
        except Exception as e:
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
            message = "An unexpected error occurred"
            print(f"An unexpected error occurred: {traceback.format_exc()}")
            self.finish(json.dumps({"error": message}))

class GetLibraryDirectoryRouteHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        library_name = input_data.get("libraryName")

        if not library_name:
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.finish(json.dumps({"error": "Library name is required"}))
            return

        directory_path = posixpath.join("xai_components", f"xai_{library_name.lower()}")
        self.finish(json.dumps({"status": "OK", "path": directory_path}))

class GetLibraryReadmeRouteHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        library_name = input_data.get("libraryName")

        if not library_name:
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.finish(json.dumps({"error": "Library name is required"}))
            return

        file_path = build_library_file_path_from_config(library_name, "readme")

        if file_path:
            if os.path.exists(file_path):
                response = {"status": "OK", "path": file_path}
            else:
                self.set_status(HTTPStatus.NOT_FOUND)
                response = {"error": "Readme file not found."}
        else:
            self.set_status(HTTPStatus.NOT_FOUND)
            response = {"error": "Readme configuration not found."}

        self.finish(json.dumps(response))

class GetLibraryExampleRouteHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        library_name = input_data.get("libraryName")

        if not library_name:
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.finish(json.dumps({"error": "Library name is required"}))
            return

        example_path = build_library_file_path_from_config(library_name, "default_example_path")

        if example_path:
            if os.path.exists(example_path):
                response = {"status": "OK", "path": example_path}
            else:
                self.set_status(HTTPStatus.NOT_FOUND)
                response = {"error": "Example file not found."}
        else:
            self.set_status(HTTPStatus.NOT_FOUND)
            response = {"error": "Example configuration not found."}

        self.finish(json.dumps(response))

class ReloadComponentLibraryConfigHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        try:
            save_component_library_config()
            response = {"status": "OK", "message": "Library config updated."}
        except Exception as e:
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
            response = {"error": f"Something went wrong: {str(e)}"}
            print(f"An error occurred: {traceback.format_exc()}")

        self.finish(json.dumps(response))


class GetComponentLibraryConfigHandler(APIHandler):
    @tornado.web.authenticated
    def get(self):
        try:
            config = get_component_library_config()
            response = {"status": "OK", "config": config}
        except Exception as e:
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
            response = {"error": f"Something went wrong: {str(e)}"}
            print(f"An error occurred: {traceback.format_exc()}")

        self.finish(json.dumps(response))

class CreateNewLibraryHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        library_name = input_data.get("libraryName", "").lower()
        component_filename = input_data.get("componentFilename", "new_component.py")
        component_content = input_data.get("componentCode", "")

        if not library_name:
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.finish(json.dumps({"error": "Library name is required"}))
            return

        try:
            message = create_or_update_library(library_name, component_filename, component_content)
            self.finish(json.dumps({"status": "OK", "message": message}))
        except Exception as e:
            self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
            message = f"An unexpected error occurred: {traceback.format_exc()}"
            print(message)
            self.finish(json.dumps({"error": message}))



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2024, TVB Widgets Team
#

import json
import os
import shutil
from datetime import datetime
import ebrains_drive
import numpy
from ebrains_drive.exceptions import DoesNotExist
from tvb.config.init.datatypes_registry import populate_datatypes_registry
from tvb.core.neocom import h5
from tvb.storage.storage_interface import StorageInterface
from tvb_ext_bucket.ebrains_drive_wrapper import BucketWrapper
from tvb_ext_bucket.exceptions import CollabAccessError
from tvbwidgets.core.auth import get_current_token

from tvbextxircuits.utils import *
from xai_components.base import xai_component, InArg, InCompArg
from xai_components.base_tvb import ComponentWithWidget
from xai_components.logger.builder import get_logger
from typing import Literal

LOGGER = get_logger(__name__)


@xai_component(color='rgb(153,0,102)')
class StoreResultsToDrive(ComponentWithWidget):
    data_to_store: InCompArg[list]
    collab_name: InCompArg[str]
    folder_path: InArg[str]
    format: InArg[Literal['h5', 'npz']]

    def __init__(self):
        self.done = False
        self.is_hpc_launch = False
        self.data_to_store = InCompArg([])
        self.collab_name = InArg('')
        self.folder_path = InArg('')
        self.format = InArg('npz')

    def execute(self, ctx) -> None:
        args = ctx.get('args')
        if args is not None:
            self.is_hpc_launch = args.is_hpc_launch
        if self.is_hpc_launch:
            # Save the config in json format for stage-out step
            json_config = {COLLAB_NAME_KEY: self.collab_name.value,
                           FOLDER_PATH_KEY: self.folder_path.value}
            with open(STORAGE_CONFIG_FILE, 'w') as f:
                json.dump(json_config, f)
        self._store_to_drive()

    def _store_to_drive(self):
        # prepare output folder
        temp_local_directory = create_temporary_dir()

        files_to_upload = list()
        for data_piece in self.data_to_store.value:
            data_file = StoreFactory.store_data_piece(data_piece, temp_local_directory, self.format.value)
            files_to_upload.append(data_file)

        if not self.is_hpc_launch:
            is_stored = self._upload_to_drive(files_to_upload, temp_local_directory)
            if is_stored:
                remove_temporary_dir(temp_local_directory)

    @staticmethod
    def connect_to_drive():
        bearer_token = get_current_token()
        client = ebrains_drive.connect(token=bearer_token)
        return client

    @staticmethod
    def create_results_folder_in_collab(collab_name, folder_path, output_directory):
        drive_client = StoreResultsToDrive.connect_to_drive()
        repos = drive_client.repos.get_repos_by_name(collab_name)

        if repos is None or len(repos) == 0:
            LOGGER.error(f'Could not find the selected Collab {collab_name} within the Drive.')
            return False

        if len(repos) > 1:
            LOGGER.warn(f'Found multiple Collabs with the name {collab_name}. Storing files to the first one...')

        if not folder_path.startswith('/'):
            folder_path = '/' + folder_path

        try:
            folder = repos[0].get_dir(folder_path)
        except DoesNotExist:
            folder = repos[0].get_dir('/')
            LOGGER.warn(f'Folder {folder_path} does not exist under the chosen Collab. Will store files under root...')

        sub_folder = folder.mkdir(output_directory)
        LOGGER.info(f'Folder {sub_folder.path} has been created into Drive, under Collab {collab_name}')
        return sub_folder

    def _upload_to_drive(self, files_to_upload, output_directory):
        LOGGER.info('Uploading files to Drive...')
        sub_folder = self.create_results_folder_in_collab(self.collab_name.value, self.folder_path.value,
                                                          output_directory)
        if sub_folder is False:
            return False

        for file_to_upload in files_to_upload:
            file = sub_folder.upload_local_file(file_to_upload)
            LOGGER.info(f'File {file.path} has been stored to Drive')
        return True


@xai_component(color='rgb(153,0,102)')
class StoreResultsToBucket(ComponentWithWidget):
    data_to_store: InCompArg[list]
    bucket_name: InCompArg[str]
    folder_path: InArg[str]
    format: InArg[Literal['h5', 'npz']]

    def __init__(self):
        self.done = False
        self.is_hpc_launch = False
        self.data_to_store = InCompArg([])
        self.bucket_name = InArg('')
        self.folder_path = InArg('')
        self.format = InArg('npz')

    def execute(self, ctx) -> None:
        args = ctx.get('args')
        if args is not None:
            self.is_hpc_launch = args.is_hpc_launch
        if self.is_hpc_launch:
            # Save the config in json format for stage-out step
            json_config = {BUCKET_NAME_KEY: self.bucket_name.value,
                           FOLDER_PATH_KEY: self.folder_path.value}
            with open(STORAGE_CONFIG_FILE, 'w') as f:
                json.dump(json_config, f)
        self._store_to_bucket()

    def _upload_to_bucket(self, files_to_upload, output_directory):
        LOGGER.info('Uploading files to Drive...')
        bucket_wrapper = BucketWrapper()

        for file_to_upload in files_to_upload:
            filename = os.path.basename(file_to_upload)
            folder = os.path.join(self.folder_path.value, output_directory)

            try:
                is_uploaded = bucket_wrapper.upload_file_to(file_to_upload, self.bucket_name.value, folder, filename)
                if is_uploaded:
                    LOGGER.info(f'File {file_to_upload} has been stored to Bucket')
                    return True
                else:
                    LOGGER.info(f'Could not upload file {file_to_upload} to the selected Bucket. '
                                f'You can find the results locally.')
                    return False

            except CollabAccessError:
                LOGGER.info(f'Could not upload file {file_to_upload} to the selected Bucket. '
                            f'You can find the results locally.')
                return False

    def _store_to_bucket(self):
        # prepare output folder
        temp_local_directory = create_temporary_dir()

        files_to_upload = list()
        for data_piece in self.data_to_store.value:
            data_file = StoreFactory.store_data_piece(data_piece, temp_local_directory, self.format.value)
            files_to_upload.append(data_file)

        if not self.is_hpc_launch:
            is_stored = self._upload_to_bucket(files_to_upload, temp_local_directory)
            if is_stored:
                remove_temporary_dir(temp_local_directory)


def create_temporary_dir():
    dirname = STORE_RESULTS_DIR + f"_{datetime.now().strftime(DIR_TIME_STAMP_FRMT)}"
    os.mkdir(dirname)
    return dirname


def remove_temporary_dir(dirname):
    LOGGER.info(f'Removing temporary local folder {dirname}')
    shutil.rmtree(dirname)


class StoreFactory(object):

    @staticmethod
    def store_data_piece(data_piece, output_directory, save_format):
        from tvb.datatypes.time_series import TimeSeries

        if issubclass(type(data_piece), TimeSeries):
            return StoreFactory.store_time_series(data_piece, output_directory, save_format)
        else:
            LOGGER.error(f"Cannot store data of type {type(data_piece)} yet. "
                         f"Please contact the development team for this.")

    @staticmethod
    def store_time_series(time_series, output_directory, save_format):
        if save_format == 'h5':
            ts_file_name = StorageInterface.FILE_NAME_STRUCTURE.format(type(time_series).__name__, time_series.gid.hex)
            ts_file_path = os.path.join(output_directory, ts_file_name)
            LOGGER.info(f"Storing timeseries for {time_series.title} monitor temporary to {ts_file_path}...")
            populate_datatypes_registry()
            h5.store(time_series, ts_file_path)
        elif save_format == 'npz':
            ts_file_path = os.path.join(output_directory, f"timeseries_{time_series.title}.npz")
            LOGGER.info(f"Storing timeseries for {time_series.title} monitor temporary to {ts_file_path}...")
            numpy.savez(ts_file_path, data=time_series.data, time=time_series.time)
        else:
            raise NotImplementedError("this format is not supported yet")
        return ts_file_path



from jupyter_server.utils import url_path_join

from .compile_xircuits import CompileXircuitsFileRouteHandler
from .components import ComponentsRouteHandler, EditXircuitsFile
from .config import RunConfigRouteHandler, SplitModeConfigHandler
from .debugger import DebuggerRouteHandler
from .spark_submit import SparkSubmitRouteHandler
from .request_library import InstallLibraryRouteHandler, FetchLibraryRouteHandler, GetLibraryDirectoryRouteHandler, GetLibraryReadmeRouteHandler, GetLibraryExampleRouteHandler, ReloadComponentLibraryConfigHandler, GetComponentLibraryConfigHandler, CreateNewLibraryHandler


def setup_handlers(web_app, url_path):
    host_pattern = ".*$"
    base_url = web_app.settings["base_url"]

    # Prepend the base_url so that it works in a JupyterHub setting
    web_app.add_handlers(host_pattern, [
        (
            url_path_join(base_url, url_path, "debug/enable"),
            DebuggerRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "config/run"),
            RunConfigRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "config/split_mode"),
            SplitModeConfigHandler
        ),
        (
            url_path_join(base_url, url_path, "components/"),
            ComponentsRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "spark/submit"),
            SparkSubmitRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "file/compile"),
            CompileXircuitsFileRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "components_edit/"),
            EditXircuitsFile
        ),
        (
            url_path_join(base_url, url_path, "library/reload_config"),
            ReloadComponentLibraryConfigHandler
        ),
        (
            url_path_join(base_url, url_path, "library/get_config"),
            GetComponentLibraryConfigHandler
        ),
        (
            url_path_join(base_url, url_path, "library/fetch"),
            FetchLibraryRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "library/install"),
            InstallLibraryRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "library/get_directory"),
            GetLibraryDirectoryRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "library/get_readme"),
            GetLibraryReadmeRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "library/get_example"),
            GetLibraryExampleRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "library/new"),
            CreateNewLibraryHandler
        )
    ])



import ast
import itertools
from collections import namedtuple
import re
import json
import sys
import os
from .port import DYNAMIC_PORTS

if sys.version_info >= (3, 9):
    from ast import unparse
else:
    import io
    from tvbextxircuits.compiler.vendor.unparse import Unparser


    def unparse(parsed):
        f = io.StringIO()
        Unparser(parsed, f)
        return f.getvalue()


def _get_value_from_literal_port(port):
    if port.source.type == "string" or port.source.type == "secret" or port.source.type == "numpy.ndarray":
        value = port.sourceLabel
    elif port.source.type == "list":
        value = json.loads("[" + port.sourceLabel + "]")
    elif port.source.type == "dict":
        value = json.loads("{" + port.sourceLabel + "}")
    elif port.source.type == "chat":
        value = json.loads(port.sourceLabel)
    elif port.source.type == "tuple":
        if port.sourceLabel == "":
            value = ()
        else:
            value = eval(port.sourceLabel)
            if not isinstance(value, tuple):
                # Handler for single entry tuple
                value = (value,)
    else:
        value = eval(port.sourceLabel)

    return value

class CodeGenerator:
    def __init__(self, graph, component_python_paths):
        self.graph = graph
        self.component_python_paths = component_python_paths

    def generate(self, filelike):
        output_filename = os.path.basename(filelike.name)
        flow_name = re.sub(r'\W', '_', output_filename.replace('.py', ''))

        module_body = list(itertools.chain(
            self._generate_python_path(),
            self._generate_fixed_imports(),
            self._generate_component_imports(),
            self._generate_flows(flow_name),
            self._generate_main(flow_name),
            self._generate_trailer()
        ))

        with open(filelike.name, 'w', encoding='utf-8') as f:
            f.write(unparse(ast.Module(body=module_body, type_ignores=[])))

    def _generate_python_path(self):
        fixed_code = """
import sys        
"""
        code = ast.parse(fixed_code).body
        nodes = self._build_node_set()

        needed_paths = set()
        for node in nodes:
            if node.name in self.component_python_paths:
                needed_paths.add(self.component_python_paths[node.name])

        if len(needed_paths) == 0:
            return []

        for p in needed_paths:
            tpl = ast.parse('sys.path.append("value")')
            tpl.body[0].value.args[0].value = p
            code.extend(tpl.body)
        return code

    def _generate_fixed_imports(self):
        fixed_imports = """
from argparse import ArgumentParser
from xai_components.base import SubGraphExecutor, InArg, OutArg, Component, xai_component, parse_bool

"""
        return ast.parse(fixed_imports).body

    def _generate_component_imports(self):
        nodes = self._build_node_set()
        unique_components = list(set((n.name, n.file) for n in nodes if n.file is not None))
        unique_components.sort(key=lambda x: x[1])
        unique_modules = itertools.groupby(unique_components, lambda x: x[1])

        imports = []
        for (file, components) in unique_modules:
            module = '.'.join(file[:-3].split('/'))

            imports.append(
                ast.parse("from %s import %s" % (module, ",".join(c[0] for c in components)))
            )

        return imports

    def _generate_flows(self, flow_name):
        mainFlowCls = ast.parse("""
@xai_component(type="xircuits_workflow")
class %s(Component):
    def __init__(self):
        super().__init__()
        self.__start_nodes__ = []
    
    def execute(self, ctx):
        pass
""" % flow_name).body[0]

        nodes = self._build_node_set()
        component_nodes = [n for n in nodes if n.file is not None]
        named_nodes = dict((n.id, "self.c_%s" % idx) for idx, n in enumerate(component_nodes))

        finish_node = [n for n in nodes if n.name == 'Finish' and n.type == 'Finish'][0]

        init_code = []
        exec_code = []
        args_code = []

        existing_args = set()

        # Instantiate all components
        init_code.extend([
            ast.parse("%s = %s()" % (named_nodes[n.id], n.name)) for n in component_nodes
        ])

        type_mapping = {
            "int": "int",
            "string": "str",
            "boolean": "bool",
            "float": "float",
            "any": "any"
        }

        def connect_args(target, source):
            return ast.parse("%s.connect(%s)" % (target, source))

        set_value = lambda target, v: ast.parse("%s.value = %s" % (target, v))

        # Set up component argument links
        for node in component_nodes:
            # Handle argument connections
            for port in (p for p in node.ports if
                         p.direction == 'in' and p.type == 'triangle-link' and p.source.name.startswith(
                             'Argument ')):
                # Unfortunately, we don't have the information anywhere else and updating the file format isn't an option at the moment

                pattern = re.compile(r'^Argument \((.+?)\): (.+)$')
                match = pattern.match(port.source.name)
                arg_type = type_mapping[match.group(1)]
                arg_name = match.group(2)

                assignment_target = "%s.%s" % (
                    named_nodes[port.target.id],
                    port.targetLabel
                )
                assignment_source = "self.%s" % arg_name
                tpl = connect_args(assignment_target, assignment_source)
                init_code.append(tpl)
                if arg_name not in existing_args:
                    in_arg_def = ast.parse("%s: InArg[%s]" % (arg_name, arg_type)).body[0]
                    args_code.append(in_arg_def)
                    existing_args.add(arg_name)

            # Handle regular connections
            for port in (p for p in node.ports if
                         p.direction == 'in' and p.type != 'triangle-link' and p.dataType not in DYNAMIC_PORTS):
                assignment_target = "%s.%s" % (
                    named_nodes[port.target.id],
                    port.targetLabel
                )

                if port.source.id not in named_nodes:
                    # Literal
                    tpl = set_value(assignment_target, '1')
                    tpl.body[0].value.value = _get_value_from_literal_port(port)
                else:
                    assignment_source = "%s.%s" % (
                        named_nodes[port.source.id],
                        port.sourceLabel
                    )
                    tpl = connect_args(assignment_target, assignment_source)
                init_code.append(tpl)

            # Handle dynamic connections
            dynaports = [p for p in node.ports if
                         p.direction == 'in' and p.type != 'triangle-link' and p.dataType in DYNAMIC_PORTS]
            ports_by_varName = {}

            RefOrValue = namedtuple('RefOrValue', ['value', 'is_ref'])  # Renamed to RefOrValue

            # Group ports by varName
            for port in dynaports:
                if port.varName not in ports_by_varName:
                    ports_by_varName[port.varName] = []
                ports_by_varName[port.varName].append(port)

            for varName, ports in ports_by_varName.items():
                dynaport_values = []

                for port in ports:
                    if port.source.id not in named_nodes:
                        value = _get_value_from_literal_port(port)
                        dynaport_values.append(RefOrValue(value, False))
                    else:
                        # Handle named node references
                        value = "%s.%s" % (named_nodes[port.source.id], port.sourceLabel)  # Variable reference
                        dynaport_values.append(RefOrValue(value, True))

                if ports[0].dataType == 'dynatuple':
                    tuple_elements = [item.value if item.is_ref else repr(item.value) for item in dynaport_values]
                    if len(tuple_elements) == 1:
                        assignment_value = '(' + tuple_elements[0] + ',)'
                    else:
                        assignment_value = '(' + ', '.join(tuple_elements) + ')'
                else:
                    list_elements = [item.value if item.is_ref else repr(item.value) for item in dynaport_values]
                    assignment_value = '[' + ', '.join(list_elements) + ']'

                assignment_target = "%s.%s" % (named_nodes[ports[0].target.id], ports[0].varName)
                tpl = set_value(assignment_target, assignment_value)
                init_code.append(tpl)

        # Handle output connections
        for i, port in enumerate(p for p in finish_node.ports if p.dataType == 'dynalist'):
            port_name = 'output'
            if i > 0:
                port_name = "%s_%s" % (port_name, i)

            assignment_target = "self.%s" % port_name
            if port.source.id not in named_nodes:
                # Literal
                tpl = set_value(assignment_target, '1')
                value = _get_value_from_literal_port(port)
                tpl.body[0].value.value = value
                port_type = type(value).__name__
            else:
                port_type = type_mapping[port.sourceType] if port.sourceType in type_mapping else port.sourceType
                assignment_source = "%s.%s" % (
                    named_nodes[port.source.id],
                    port.sourceLabel
                )
                tpl = connect_args(assignment_target, assignment_source)
            args_code.append(ast.parse("%s: OutArg[%s]" % (port_name, port_type)).body[0])
            init_code.append(tpl)

        # Set up control flow
        for node in component_nodes:
            has_next = False
            for port in (p for p in node.ports if p.direction == "out" and p.type == "triangle-link"):
                if port.name == "out-0":
                    has_next = True
                    assignment_target = "%s.next" % named_nodes[port.source.id]
                    assignment_source = named_nodes[port.target.id] if port.target.id in named_nodes else None
                    init_code.append(
                        ast.parse("%s = %s" % (assignment_target, assignment_source))
                    )
                elif port.name.startswith("out-flow-"):
                    assignment_target = "%s.%s" % (named_nodes[port.source.id], port.name[len("out-flow-"):])
                    assignment_source = "SubGraphExecutor(%s)" % named_nodes[
                        port.target.id] if port.target.id in named_nodes else None
                    init_code.append(
                        ast.parse("%s = %s" % (assignment_target, assignment_source))
                    )
            if not has_next:
                assignment_target = "%s.next" % named_nodes[node.id]
                init_code.append(
                    ast.parse("%s = None" % assignment_target)
                )
        # Setup start node initialization
        for node in (n for n in self.graph if n.id in named_nodes):
            init_code.append(
                ast.parse("self.__start_nodes__.append(%s)" % named_nodes[node.id])
            )

        trailer = """
for node in self.__start_nodes__:
    if hasattr(node, 'init'):
        node.init(ctx)
    
next_component = %s
while next_component is not None:
    next_component = next_component.do(ctx)        
        """ % (named_nodes[self.graph[0].ports[0].target.id])
        exec_code.append(ast.parse(trailer))

        mainFlowCls.body[0].body.extend(init_code)
        mainFlowCls.body[1].body = exec_code

        args_code.sort(key=lambda x: x.target.id)
        mainFlowCls.body = args_code + mainFlowCls.body
        return [mainFlowCls]

    def _generate_main(self, flow_name):
        main = ast.parse("""
def main(args):
    import pprint
    ctx = {}
    ctx['args'] = args
    flow = %s()
    flow.next = None
""" % flow_name).body[0]
        pattern = re.compile(r'^Argument \(.+?\): (.+)$')

        body = main.body

        # Set up the input values
        nodes = self._build_node_set()
        finish_node = [n for n in nodes if n.name == 'Finish' and n.type == 'Finish'][0]
        argument_nodes = [n for n in nodes if n.name.startswith("Argument ") and n.file is None]
        for arg in argument_nodes:
            m = pattern.match(arg.name)
            arg_name = m.group(1)
            tpl = "flow.%s.value = args.%s" % (arg_name, arg_name)
            body.extend(ast.parse(tpl).body)

        body.extend(ast.parse("""
flow.do(ctx)
""").body)

        # Print out the output values
        for i, port in enumerate(p for p in finish_node.ports if p.dataType == 'dynalist'):
            port_name = 'output'
            if i > 0:
                port_name = "%s_%s" % (port_name, i)

            body.extend(ast.parse("""
print("%s:")
pprint.pprint(flow.%s.value)
""" % (port_name, port_name)).body)

        return [main]

    def _build_node_set(self):
        nodes = set()
        node_queue = list(self.graph)
        while len(node_queue) != 0:
            current_node = node_queue.pop()
            nodes.add(current_node)
            for port in current_node.ports:
                if port.target not in nodes:
                    node_queue.append(port.target)
                if port.source not in nodes:
                    node_queue.append(port.source)
        return nodes

    def _generate_trailer(self):
        code = """
if __name__ == '__main__':
    args, _ = parser.parse_known_args()
    main(args)
    print("\\nFinished Executing")
        """
        body = ast.parse(code).body[0]
        arg_parsing = self._generate_argument_parsing()
        arg_parsing.extend(body.body)
        body.body = arg_parsing
        return [body]

    def _generate_argument_parsing(self):
        # Unfortunately, we don't have the information anywhere else and updating the file format isn't an option at the moment
        pattern = re.compile(r'^Argument \(.+?\): (.+)$')

        type_mapping = {
            "int": "int",
            "string": "str",
            "float": "float",
            "any": "any"
        }

        code = """
parser = ArgumentParser()
parser.add_argument('--is_hpc_launch', default=False, type=bool)  
        """
        body = ast.parse(code).body

        nodes = self._build_node_set()
        argument_nodes = [n for n in nodes if n.name.startswith("Argument ") and n.file is None]
        for arg in argument_nodes:
            m = pattern.match(arg.name)
            arg_name = m.group(1)
            if arg.type == "boolean":
                tpl = "parser.add_argument('--%s', type=parse_bool, default=None, nargs='?', const=True)" % arg_name
            else:
                tpl = "parser.add_argument('--%s', type=%s)" % (arg_name, type_mapping[arg.type])
            body.extend(ast.parse(tpl).body)

        return body


from xai_components.base import InArg, OutArg, InCompArg, Component, xai_component, dynalist, dynatuple, BaseComponent, SubGraphExecutor

import os
import sys
from pathlib import Path
import time
import datetime
import json
import random

@xai_component
class GetCurrentTime(Component):
    """Retrieves the current time in ISO 8601 format.

    ##### outPorts:
    - time_str (str): The current time as a string in ISO 8601 format.
    """
    time_str: OutArg[str]

    def execute(self, ctx) -> None:
        try:
            import pytz

            tz = pytz.timezone('UTC')
            now = datetime.datetime.now(tz)
        except:
            now = datetime.datetime.now()

        self.time_str.value = now.isoformat()

@xai_component
class GetCurrentDate(Component):
    """Retrieves the current date components.

    ##### outPorts:
    - year (str): The current year.
    - month (str): The current month.
    - day (str): The current day.
    """
    year: OutArg[str]
    month: OutArg[str]
    day: OutArg[str]

    def execute(self, ctx) -> None:
        today = datetime.date.today()

        self.year.value = str(today.year)
        self.month.value = str(today.month)
        self.day.value = str(today.day)

@xai_component
class Print(Component):
    """Prints a message to the console.

    ##### inPorts:
    - msg (any): The message to be printed.
    """
    msg: InArg[any]

    def execute(self, ctx) -> None:
        print(str(self.msg.value))

@xai_component
class PrettyPrint(Component):
    """Prints a message in a pretty format using pprint.

    ##### inPorts:
    - msg (any): The message to be pretty printed.
    """
    msg: InArg[any]

    def execute(self, ctx) -> None:
        import pprint
        pp = pprint.PrettyPrinter(indent=4)
        print(pp.pformat(self.msg.value))

@xai_component
class ConcatString(Component):
    """Concatenates two strings.

    ##### inPorts:
    - a (str): The first string.
    - b (str): The second string.

    ##### outPorts:
    - out (str): The concatenated result of strings a and b.
    """
    a: InArg[str]
    b: InArg[str]
    out: OutArg[str]

    def execute(self, cts) -> None:
        self.out.value = self.a.value + self.b.value

@xai_component
class FormatString(Component):
    """Formats a string using placeholders and a dictionary of replacements.

    ##### inPorts:
    - format_str (str): The string with placeholders.
    - args (dict): Dictionary with placeholder replacements.

    ##### outPorts:
    - out_str (str): The formatted string.

    ##### Example:
    format_str: "Hello, {name}! Today is {day}."
    args: {"name": "Xircuits", "day": "Monday"}
    out_str: "Hello, Xircuits! Today is Monday."
    """
    format_str: InCompArg[str]
    args: InArg[dict]
    out_str: OutArg[str]

    def execute(self, ctx):
        self.out_str.value = self.format_str.value.format(**self.args.value)

@xai_component
class SplitString(Component):
    """Splits a string by a specified character.

    ##### inPorts:
    - string (str): The string to split.
    - ch (str): The character to split the string by.

    ##### outPorts:
    - out (list): List of substrings.
    """
    string: InArg[str]
    ch: InArg[str]
    out: OutArg[list]

    def execute(self, cts) -> None:
        self.out.value = self.string.value.split(self.ch.value)

@xai_component
class JoinArrayWithString(Component):
    """Joins an array of strings into a single string with a separator.

    ##### inPorts:
    - array (list): The array of strings to join.
    - sep (str): The separator string.

    ##### outPorts:
    - out (str): The joined string.
    """
    array: InArg[list]
    sep: InArg[str]
    out: OutArg[str]

    def execute(self, cts) -> None:
        self.out.value = self.sep.value.join(self.array.value)

@xai_component
class ZipDirectory(Component):
    """Zips a directory.

    ##### inPorts:
    - zip_fn (str): The Zip filename. Default: .xircuits canvas name.
    - dir_name (str): The directory to be zipped.
    - include_dir (bool): Bundle the specified directory in the zip if True.

        #### EG:
        Creating Bar.zip which zips the dir `Foo` with the structure:
        ```
        Foo
        |-- file1
        |-- file2
        ```

        if `include_dir` is `True`:
        ```
        Bar.zip
        |-- Foo
            |-- file1
            |-- file2
        ```

        if `include_dir` is `False`:
        ```
        Bar.zip
        |-- file1
        |-- file2
        ```

    """
    zip_fn: InArg[str]
    dir_name: InCompArg[str]
    include_dir: InArg[bool]

    def execute(self, ctx) -> None:
        from zipfile import ZipFile
        from tqdm import tqdm

        zip_fn = self.zip_fn.value if self.zip_fn.value else Path(sys.argv[0]).stem
        dir_name = self.dir_name.value

        assert os.path.isdir(dir_name), "invalid dir_name!"

        if os.path.splitext(zip_fn)[-1] != ".zip":
            zip_fn = zip_fn + ".zip"

        if not Path(zip_fn).is_file():
            print(zip_fn + " created at " + os.getcwd())
            zipObj = ZipFile(zip_fn, 'w')
        else:
            print(zip_fn + " updated at " + os.getcwd())
            zipObj = ZipFile(zip_fn,'a')

        for root, dirs, files in tqdm(list(os.walk(dir_name))):
            # chop off root dir
            if self.include_dir.value == False:
                length = len(dir_name)
                dirs = root[length:]
            for filename in files:
                if self.include_dir.value == False:
                    zipObj.write(os.path.join(root, filename), os.path.join(dirs, filename))
                else:
                    zipObj.write(os.path.join(root, filename))

        zipObj.close()

@xai_component
class MoveFile(Component):
    """Move a file.

    ##### inPorts:
    - source_path (str): Path to the file to be moved.
    - dest_path (str): The destination.

    ##### outPorts:
    - result_path (str): Resulting file path.
    """
    source_path: InArg[str]
    dest_path: InArg[str]
    result_path: OutArg[str]

    def execute(self, ctx) -> None:
        import shutil
        new_path = shutil.move(self.source_path.value, self.dest_path.value)
        self.result_path.value = new_path

@xai_component
class CopyFile(Component):
    """Copies a file.

    ##### inPorts:
    - source_path (str): Path to the file to be copied.
    - dest_path (str): The destination.

    ##### outPorts:
    - result_path (str): Resulting file path.
    """
    source_path: InArg[str]
    dest_path: InArg[str]
    result_path: OutArg[str]

    def execute(self, ctx) -> None:
        import shutil
        new_path = shutil.copy2(self.source_path.value, self.dest_path.value)
        self.result_path.value = new_path

@xai_component
class DeleteFile(Component):
    """Deletes a file.

    ##### inPorts:
    - filename (str): Path to file to be deleted.
    """
    filename: InCompArg[str]

    def execute(self, ctx) -> None:
        filename = self.filename.value if self.filename.value else None
        if os.path.exists(filename):
          os.remove(filename)
        else:
          print(filename + " does not exist.")

@xai_component(color="green")
class TimerComponent(Component):
    """Chain multiple instances of this component to measure elapsed time.

    ##### inPorts:
    - in_timer (time.time): If provided will measure the elapsed time since last called. If not provided, will start a new timer.
    - timer_message (str): A log message to be printed.

    ##### outPorts:
    - out_timer (time.time): Passes current timer to the next TimerComponent.
    - elapsed_time (int): The elapsed time in seconds.
    """
    in_timer: InArg[time.time]
    timer_message: InArg[str]

    out_timer: OutArg[time.time]
    elapsed_time: OutArg[int]

    def execute(self, ctx) -> None:
        current_time = time.time()
        timer_message = self.timer_message.value if self.timer_message.value else "current timer component"

        elapsed_time = 0
        if self.in_timer.value:
                elapsed_time = current_time - self.in_timer.value
                print("The elapsed time for " + timer_message + " is " + str(elapsed_time) + " seconds.")
        else:
            print("Starting new timer.")

        self.out_timer.value = current_time
        self.elapsed_time.value = elapsed_time

@xai_component(color="green")
class SleepComponent(Component):
    """Pauses the python process.

    ##### inPorts:
    - sleep_timer (float): The number of seconds to pause. Default `5.0` seconds.
    """
    sleep_timer: InArg[float]

    def execute(self, ctx) -> None:
        sleep_timer = self.sleep_timer.value if self.sleep_timer.value else 5.0
        print("Sleeping for " + str(sleep_timer) + " seconds.")
        time.sleep(sleep_timer)

@xai_component(color="grey")
class MakeList(Component):
    """A component that takes values from its dynamic list port and outputs as a normal list.

    ##### inPorts:
    - list_values (dynalist): Dynamic list port that can take any vars and append it in a list.

    ##### outPorts:
    - output_list (list): The constructed list from the dynamic list inPorts.
    """
    list_values: InArg[dynalist]
    output_list: OutArg[list]

    def execute(self, ctx) -> None:
        self.output_list.value = self.list_values.value
        print("Constructed List:", self.output_list.value)

@xai_component(color="grey")
class MakeTuple(Component):
    """A component that takes values from its dynamic tuple port and outputs as a normal tuple.

    ##### inPorts:
    - tuple_values (dynatuple): Dynamic tuple port that can take any vars and append it in a tuple.

    ##### outPorts:
    - output_tuple (tuple): The constructed tuple from the dynamic tuple inPorts.
    """
    tuple_values: InArg[dynatuple]
    output_tuple: OutArg[tuple]

    def execute(self, ctx) -> None:
        self.output_tuple.value = self.tuple_values.value
        print("Constructed Tuple:", self.output_tuple.value)

@xai_component(color="grey")
class MakeDict(Component):
    """A component that takes two dynamic lists (dynalists) as inputs - one for keys and one for values,
    and constructs a dictionary from these lists. If there are more keys than values, the unmatched keys
    will have None as their value.

    ##### inPorts:
    - keys_list (dynalist): Dynamic list of keys for the dictionary.
    - values_list (dynalist): Dynamic list of values for the dictionary.

    ##### outPorts:
    - output_dict (dict): The constructed dictionary from the provided keys and values.
    """
    keys_list: InArg[dynalist]
    values_list: InArg[dynalist]
    output_dict: OutArg[dict]

    def execute(self, ctx) -> None:
        keys = self.keys_list.value
        values = self.values_list.value

        constructed_dict = {key: values[i] if i < len(values) else None for i, key in enumerate(keys)}

        self.output_dict.value = constructed_dict
        print("Constructed Dictionary:", self.output_dict.value)

@xai_component(color="orange")
class ExecuteNotebook(Component):
    """Executes a Jupyter notebook from a given file path and logs the output.

    This component runs a Jupyter notebook and can save a log of the execution.
    It's useful for automated running of notebooks for data analysis or similar tasks.

    ##### inPorts:
    - notebook_filepath (str): Path of the '.ipynb' file to execute.
    - log_filepath (str): Path for saving the execution log (optional).

    ##### Notes:
    - Only '.ipynb' files are accepted.
    - Execution timeout is 10 minutes.
    - Errors during execution are logged or printed.
    """
    notebook_filepath: InCompArg[str]
    log_filepath: InArg[str]

    def execute(self, context):
        import nbformat
        from nbconvert.preprocessors import ExecutePreprocessor, CellExecutionError

        notebook_filepath = self.notebook_filepath.value

        # Ensure that the input file has a valid ".ipynb" extension
        if not notebook_filepath.lower().endswith(".ipynb"):
            raise ValueError("Invalid input file path. Please provide a valid jupyter notebook '.ipynb' file.")

        # Extract the filename and directory path
        input_filename, _ = os.path.splitext(os.path.basename(notebook_filepath))
        log_filepath = self.log_filepath.value

        with open(notebook_filepath) as f:
            nb = nbformat.read(f, as_version=4)

        ep = ExecutePreprocessor(timeout=600, kernel_name='python3')

        try:
            out = ep.preprocess(nb, {'metadata': {'path': '.'}})
        except CellExecutionError:
            msg = 'Error executing the notebook "%s".\n\n' % notebook_filepath
            msg += 'See notebook "%s" for the traceback.' % log_filepath
            print(msg)
            log_filepath if log_filepath else f"{input_filename}-output.ipynb"
            raise
        finally:
            if log_filepath:
                with open(log_filepath, mode='w', encoding='utf-8') as f:
                    nbformat.write(nb, f)

@xai_component
class IsNone(Component):
    """Checks if the input value is None.

    ##### inPorts:
    - a (any): The value to be checked.

    ##### outPorts:
    - out (bool): True if 'a' is None, False otherwise.
    """
    a: InArg[any]
    out: OutArg[bool]

    def execute(self, ctx) -> None:
        self.out.value = self.a.value is None

@xai_component
class IsNotNone(Component):
    """Checks if the input value is not None.

    ##### inPorts:
    - a (any): The value to be checked.

    ##### outPorts:
    - out (bool): True if 'a' is not None, False otherwise.
    """
    a: InArg[any]
    out: OutArg[bool]

    def execute(self, ctx) -> None:
        self.out.value = self.a.value is not None

@xai_component
class SetDictValue(Component):
    """Sets a key-value pair in a dictionary, creating a new one if none exists.

    ##### inPorts:
    - dict (dict): Input or existing dictionary.
    - key (str): Key to set in the dictionary.
    - value (any): Value to set for the key.

    ##### outPorts:
    - out_dict (dict): Dictionary with the updated key-value pair.
    """
    dict: InArg[dict]
    key: InArg[str]
    value: InArg[any]
    out_dict: OutArg[dict]

    def execute(self, ctx) -> None:
        self.out_dict.value = {} if self.dict.value is None else self.dict.value
        self.out_dict.value[self.key.value] = self.value.value

@xai_component
class GetDictValue(Component):
    """Retrieves a value from a dictionary by key.

    ##### inPorts:
    - dict (dict): The dictionary to search.
    - key (str): The key for the value to retrieve.

    ##### outPorts:
    - value (any): The retrieved value, or None if the key is not found.
    """
    dict: InArg[dict]
    key: InArg[str]
    value: OutArg[any]

    def execute(self, ctx) -> None:
        self.value.value = self.dict.value.get(self.key.value)

@xai_component
class ListAppend(Component):
    """Appends an item to a list.

    ##### inPorts:
    - the_list (list): The list to which the item will be appended.
    - item (any): The item to append to the list.

    ##### outPorts:
    - out_list (list): The list with the appended item.
    """
    the_list: InArg[list]
    item: InCompArg[any]
    out_list: OutArg[list]

    def execute(self, ctx) -> None:
        l = [] if self.the_list.value is None else self.the_list.value
        l.append(self.item.value)
        self.out_list.value = l

@xai_component
class ListGetItem(Component):
    """Retrieves an item from a list based on the specified index.

    ##### inPorts:
    - the_list (list): The list from which to retrieve the item.
    - index (int): The index of the item in the list.

    ##### outPorts:
    - out_item (any): The retrieved item.
    """
    the_list: InCompArg[list]
    index: InCompArg[int]
    out_item: OutArg[any]

    def execute(self, ctx) -> None:
        self.out_item.value = self.the_list.value[self.index.value]

@xai_component
class ListSetItem(Component):
    """Sets an item in a list at the specified index.

    ##### inPorts:
    - the_list (list): The list in which to set the item.
    - index (int): The index at which to set the item.
    - item (any): The item to set in the list.

    ##### outPorts:
    - out_list (list): The list with the set item.
    """
    the_list: InCompArg[list]
    index: InCompArg[int]
    item: InCompArg[any]
    out_list: OutArg[list]

    def execute(self, ctx) -> None:
        self.the_list.value[self.index.value] = self.item.value
        self.out_list.value = self.the_list.value

@xai_component
class DictGetItem(Component):
    """Retrieves an item from a dictionary using the specified key.

    ##### inPorts:
    - the_dict (dict): The dictionary from which to retrieve the item.
    - key (any): The key corresponding to the item in the dictionary.

    ##### outPorts:
    - out_item (any): The retrieved item.
    """
    the_dict: InCompArg[dict]
    key: InCompArg[any]
    out_item: OutArg[any]

    def execute(self, ctx) -> None:
        self.out_item.value = self.the_dict.value[self.key.value]

@xai_component
class DictSetItem(Component):
    """Sets an item in a dictionary with the specified key.

    ##### inPorts:
    - the_dict (dict): The dictionary in which to set the item.
    - key (any): The key under which to set the item.
    - item (any): The item to be set in the dictionary.

    ##### outPorts:
    - out_dict (dict): The dictionary with the set item.
    """
    the_dict: InArg[dict]
    key: InCompArg[any]
    item: InCompArg[any]
    out_dict: OutArg[dict]

    def execute(self, ctx) -> None:
        d = {} if self.the_dict.value is None else self.the_dict.value
        d[self.key.value] = self.item.value
        self.out_dict.value = d

@xai_component
class ToJson(Component):
    """Converts an object to a JSON string.

    ##### inPorts:
    - obj (any): The object to convert to JSON.

    ##### outPorts:
    - json_str (str): The JSON string representation of the object.
    """
    obj: InCompArg[any]
    json_str: OutArg[str]

    def execute(self, ctx) -> None:
        self.json_str.value = json.dumps(self.obj.value)

@xai_component
class FromJson(Component):
    """Converts a JSON string to an object.

    ##### inPorts:
    - json_str (str): The JSON string to convert to an object.

    ##### outPorts:
    - obj (any): The object represented by the JSON string.
    """
    json_str: InCompArg[str]
    obj: OutArg[any]

    def execute(self, ctx) -> None:
        self.obj.value = json.loads(self.json_str.value)

@xai_component
class GetRandomNumber(Component):
    """Generates a random number between the specified bounds.

    ##### inPorts:
    - greater_than (int): The lower bound for the random number (inclusive).
    - less_than (int): The upper bound for the random number (exclusive).

    ##### outPorts:
    - value (int): The generated random number.
    """
    greater_than: InCompArg[int]
    less_than: InCompArg[int]
    value: OutArg[int]

    def execute(self, ctx) -> None:
        self.value.value = random.randint(self.greater_than.value, self.less_than.value)

@xai_component(color='blue')
class RunParallelThread(Component):
    """Executes a given body in a separate thread using a thread pool executor.

    **Important Note**: Changes done in the body are not propagated!
    This includes things like setting variable values, modifying the context, or
    even setting output arguments. This means, you can't pull the result of a
    component that is run on a separate thread.

    ##### inPorts:
    - n_workers (int): The number of worker threads to use for executing the body in parallel.

    ##### outPorts:
    - futures (list): All futures created by this component.

    ##### Branches:
    - body: The body to be executed in parallel.
    """
    n_workers: InArg[int]
    futures: OutArg[list]
    body: BaseComponent

    def __init__(self):
        super().__init__()
        self.executor = None
        self.futures.value = []

    def execute(self, ctx) -> None:
        from concurrent.futures import ThreadPoolExecutor
        from copy import deepcopy

        if self.executor is None:
            self.executor = ThreadPoolExecutor(max_workers=self.n_workers.value)

        def execute_body(body, ctx):
            SubGraphExecutor(body).do(ctx)

        x = self.executor.submit(execute_body, deepcopy(self.body), deepcopy(ctx))

        # Enforce that any exceptions are logged
        x.add_done_callback(lambda x: x.result())

        self.futures.value.append(x)

@xai_component(color='blue')
class AwaitFutures(Component):
    """Waits for a list of futures to complete.

    ##### inPorts:
    - futures (list): The list of futures to wait for.
    """
    futures: InCompArg[list]

    def execute(self, ctx) -> None:
        from concurrent.futures import wait
        wait(self.futures.value)



from tvbextxircuits.handlers.component_parser import ComponentsParser


def main():
    route_handler = ComponentsParser()
    route_handler.generate_doc_files()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# "TheVirtualBrain - Widgets" package
#
# (c) 2022-2024, TVB Widgets Team
#

import json
import os
import sys
from datetime import datetime
from io import BytesIO
from urllib.error import HTTPError
import pyunicore.client as unicore_client
import requests
from pyunicore.client import JobStatus as unicore_status
from pyunicore.credentials import AuthenticationFailedException, OIDCToken
from tvb_ext_bucket.ebrains_drive_wrapper import BucketWrapper
from tvb_ext_bucket.exceptions import CollabAccessError
from tvbwidgets.core.auth import get_current_token

import tvbextxircuits._version as xircuits_version
from tvbextxircuits.hpc_config.parse_files import get_files_to_upload
from tvbextxircuits.logger.builder import get_logger
from tvbextxircuits.utils import *
from xai_components.xai_storage.store_results import StoreResultsToDrive

LOGGER = get_logger('tvbextxircuits.hpc_config.pyunicore_config')


class PyunicoreSubmitter(object):
    storage_name = {'JUWELS': 'PROJECT', 'JUDAC': 'PROJECT'}
    env_dir = 'tvb_xircuits'
    env_name = 'venv'
    python_dir = {'JUWELS': 'python3.11', 'JUDAC': 'python3.10'}
    modules = {'JUWELS': 'Python', 'JUDAC': 'Python'}
    pip_libraries = 'tvb-ext-xircuits tvb-data'
    EXECUTABLE_KEY = 'Executable'
    PROJECT_KEY = 'Project'
    JOB_TYPE_KEY = 'Job type'
    INTERACTIVE_KEY = 'interactive'

    def __init__(self, site, project):
        self.site = site
        self.project = project

    def set_hpc_settings(self, filesystem, python, libraries, modules):
        if filesystem:
            self.storage_name[self.site] = filesystem
        if python:
            self.python_dir[self.site] = python
        if libraries:
            self.pip_libraries = libraries
        if modules:
            self.modules[self.site] = modules

    @property
    def _activate_command(self):
        return f'source ${self.storage_name[self.site]}/{self.env_dir}/{self.env_name}/bin/activate'

    @property
    def _module_load_command(self):
        return f'module load {self.modules.get(self.site, "")}'

    @property
    def _create_env_command(self):
        return f'cd ${self.storage_name[self.site]}/{self.env_dir} ' \
               f'&& rm -rf {self.env_name} ' \
               f'&& python -mvenv {self.env_name}'

    @property
    def _install_dependencies_command(self):
        return f'pip install -U pip && pip install {self.pip_libraries}'

    def connect_client(self):
        LOGGER.info(f"Connecting to {self.site}...")
        token = OIDCToken(get_current_token())
        transport = unicore_client.Transport(token)
        registry = unicore_client.Registry(transport, unicore_client._HBP_REGISTRY_URL)

        try:
            sites = registry.site_urls
        except Exception:
            LOGGER.error(f"Unicore seems to be down at the moment. "
                         f"Please check service availability and try again later")
            return None

        try:
            site_url = sites[self.site]
        except KeyError:
            LOGGER.error(f'Site {self.site} seems to be down for the moment.')
            return None

        try:
            client = unicore_client.Client(transport, site_url)
        except (AuthenticationFailedException, HTTPError):
            LOGGER.error(f'Authentication to {self.site} failed, you might not have permissions to access it.')
            return None

        LOGGER.info(f'Authenticated to {self.site} with success.')
        return client

    def _check_environment_ready(self, home_storage):
        # Pyunicore listdir method returns directory names suffixed by '/'
        if f"{self.env_dir}/" not in home_storage.listdir():
            home_storage.mkdir(self.env_dir)
            LOGGER.info(f"Environment directory not found in HOME, will be created.")
            return False

        if f"{self.env_dir}/{self.env_name}/" not in home_storage.listdir(self.env_dir):
            LOGGER.info(f"Environment not found in HOME, will be created.")
            return False

        try:
            # Check whether tvb-ext-xircuits is installed in HPC env and if version is updated
            site_packages_path = f'{self.env_dir}/{self.env_name}/lib/{self.python_dir[self.site]}/site-packages'
            site_packages = home_storage.listdir(site_packages_path)
            files = [file for file in site_packages if "tvb_ext_xircuits" in file]
            assert len(files) >= 1
            remote_version = files[0].split("tvb_ext_xircuits-")[1].split('.dist-info')[0]
            local_version = xircuits_version.__version__
            if remote_version != local_version:
                LOGGER.info(f"Found an older version {remote_version} of tvb-ext-xircuits installed in the "
                            f"environment, will recreate it with {local_version}.")
                return False
            return True
        except HTTPError as e:
            LOGGER.info(f"Could not find site-packages in the environment, will recreate it: {e}")
            return False
        except AssertionError:
            LOGGER.info(f"Could not find tvb-ext-xircuits installed in the environment, will recreate it.")
            return False
        except IndexError:
            LOGGER.info(f"Could not find tvb-ext-xircuits installed in the environment, will recreate it.")
            return False

    def _search_for_home_dir(self, client):
        LOGGER.info(f"Accessing storages on {self.site}...")
        num = 10
        offset = 0
        storages = client.get_storages(num=num, offset=offset)
        while len(storages) > 0:
            for storage in storages:
                if storage.resource_url.endswith(self.storage_name[self.site]):
                    return storage
            offset += num
            storages = client.get_storages(num=num, offset=offset)
        return None

    def _format_date_for_job(self, job):
        date = datetime.strptime(job.properties['submissionTime'], '%Y-%m-%dT%H:%M:%S+%f')
        return date.strftime('%m.%d.%Y, %H_%M_%S')

    def _dev_mode(self, home_storage, client):
        """
        The purpose of this method is to allow developers to test packages on HPC before releasing them on Pypi.
        First step is to run the build command under the tvb-ext-xircuits folder: python -m build
        It will generate the 'dist' folder with a WHL and TAR.GZ packages for tvb-ext-xircuits.
        Then, make sure to call this method from submit_job method before launching the workflow job.
        """
        local_package_name = f'tvb_ext_xircuits-{xircuits_version.__version__}-py3-none-any.whl'
        LOGGER.info(f"You are running in dev mode, starting to install {local_package_name} on HPC {self.site}...")
        home_storage.rm(local_package_name)
        home_storage.upload(
            input_file=f'dist/{local_package_name}',
            destination=f'{self.env_dir}/{local_package_name}')
        self.pip_libraries = self.pip_libraries.replace('tvb-ext-xircuits', local_package_name)
        job_description = {
            self.EXECUTABLE_KEY: f"{self._module_load_command} && {self._create_env_command} && "
                                 f"{self._activate_command} && {self._install_dependencies_command}",
            self.PROJECT_KEY: self.project,
            self.JOB_TYPE_KEY: self.INTERACTIVE_KEY}
        job_env_prep = client.new_job(job_description, inputs=[])
        LOGGER.info(f"Job is running at {self.site}: {job_env_prep.working_dir.properties['mountPoint']}. "
                    f"Submission time is: {self._format_date_for_job(job_env_prep)}. "
                    f"Waiting for job to finish..."
                    f"It can also be monitored interactively with the Monitor HPC button.")
        job_env_prep.poll()
        if job_env_prep.properties['status'] == unicore_status.FAILED:
            LOGGER.error(f"Encountered an error during environment setup, stopping execution.")
            return
        LOGGER.info(f"Successfully finished the environment setup.")

    def submit_job(self, executable, inputs, do_stage_out):
        client = self.connect_client()
        if client is None:
            LOGGER.error(f"Could not connect to {self.site}, stopping execution.")
            return

        home_storage = self._search_for_home_dir(client)
        if home_storage is None:
            LOGGER.error(f"Could not find a {self.storage_name[self.site]} storage on {self.site}, stopping execution.")
            return

        is_env_ready = self._check_environment_ready(home_storage)

        if is_env_ready:
            LOGGER.info(f"Environment is already prepared, it won't be recreated.")
            # self._dev_mode(home_storage, client)
        else:
            LOGGER.info(f"Preparing environment in your {self.storage_name[self.site]} folder...")
            job_description = {
                self.EXECUTABLE_KEY: f"{self._module_load_command} && {self._create_env_command} && "
                                     f"{self._activate_command} && {self._install_dependencies_command}",
                self.PROJECT_KEY: self.project,
                self.JOB_TYPE_KEY: self.INTERACTIVE_KEY}
            job_env_prep = client.new_job(job_description, inputs=[])
            LOGGER.info(f"Job is running at {self.site}: {job_env_prep.working_dir.properties['mountPoint']}. "
                        f"Submission time is: {self._format_date_for_job(job_env_prep)}. "
                        f"Waiting for job to finish..."
                        f"It can also be monitored interactively with the Monitor HPC button.")
            job_env_prep.poll()
            if job_env_prep.properties['status'] == unicore_status.FAILED:
                LOGGER.error(f"Encountered an error during environment setup, stopping execution.")
                return
            LOGGER.info(f"Successfully finished the environment setup.")

        LOGGER.info("Launching workflow...")
        job_description = {
            self.EXECUTABLE_KEY: f"{self._module_load_command} && {self._activate_command} && "
                                 f"python {executable} --is_hpc_launch=True",
            self.PROJECT_KEY: self.project}
        job_workflow = client.new_job(job_description, inputs=inputs)
        LOGGER.info(f"Job is running at {self.site}: {job_workflow.working_dir.properties['mountPoint']}. "
                    f"Submission time is: {self._format_date_for_job(job_workflow)}.")
        LOGGER.info('Finished remote launch.')

        if do_stage_out:
            self.monitor_job(job_workflow)

        else:
            LOGGER.info('You can use Monitor HPC button to monitor it.')

    def monitor_job(self, job):
        LOGGER.info('Waiting for job to finish...'
                    'It can also be monitored interactively with the Monitor HPC button.')
        job.poll()

        if job.properties['status'] == unicore_status.FAILED:
            LOGGER.error(f"Job finished with errors.")
            return
        LOGGER.info(f"Job finished with success. Staging out the results...")
        self.stage_out_results(job)
        LOGGER.info(f"Finished execution.")

    def stage_out_results(self, job):
        content = job.working_dir.listdir()

        results_dirname = None
        for file_name in content.keys():
            if file_name.startswith(STORE_RESULTS_DIR):
                results_dirname = file_name

        if results_dirname is None:
            LOGGER.info(f"Could not find results folder for this job. Nothing to stage out.")
            return

        LOGGER.info(f"Found sub dir: {results_dirname}")
        results_content = job.working_dir.listdir(results_dirname)

        storage_config_file = content.get(STORAGE_CONFIG_FILE)
        if storage_config_file is None:
            LOGGER.info(f"Could not find file: {STORAGE_CONFIG_FILE}")
            LOGGER.info("Could not finalize the stage out. "
                        "Please download your results manually using the Monitor HPC button.")
            return
        else:
            storage_config_file.download(STORAGE_CONFIG_FILE)
            with open(STORAGE_CONFIG_FILE) as f:
                storage_config = json.load(f)
            os.remove(STORAGE_CONFIG_FILE)

        collab_name = storage_config.get(COLLAB_NAME_KEY)
        bucket_name = storage_config.get(BUCKET_NAME_KEY)
        folder_path = storage_config.get(FOLDER_PATH_KEY)

        if bucket_name is None:
            self._stage_out_results_to_drive(results_content, collab_name, folder_path, results_dirname)
        else:
            self._stage_out_results_to_bucket(results_content, bucket_name, folder_path, results_dirname)

    def _stage_out_results_to_drive(self, results_folder_content, collab_name, folder_path, results_dirname):
        LOGGER.info(f"Storing results to Collab {collab_name} under {folder_path}/{results_dirname} ...")
        sub_folder = StoreResultsToDrive.create_results_folder_in_collab(collab_name, folder_path, results_dirname)

        for key, val in results_folder_content.items():
            if isinstance(val, unicore_client.PathFile):
                with BytesIO() as in_memory_file:
                    val.download(in_memory_file)
                    file = sub_folder.upload(in_memory_file.getvalue(), os.path.basename(key))
                    LOGGER.info(f'File {file.path} has been stored to Drive')

    def _stage_out_results_to_bucket(self, results_folder_content, bucket_name, folder_path, results_dirname):
        LOGGER.info(f"Storing results to Bucket {bucket_name} under {folder_path}/{results_dirname}")
        bucket_wrapper = BucketWrapper()

        for key, val in results_folder_content.items():
            if isinstance(val, unicore_client.PathFile):
                with BytesIO() as in_memory_file:
                    val.download(in_memory_file)

                    try:
                        upload_url = bucket_wrapper.get_bucket_upload_url(bucket_name, os.path.basename(key),
                                                                          os.path.join(folder_path, results_dirname))
                    except CollabAccessError:
                        LOGGER.info(f'Could not upload file {key} to the selected Bucket. '
                                    f'You can find the results under the job directory.')
                        return
                    resp = requests.request("PUT", upload_url, data=in_memory_file.getvalue())
                    resp.raise_for_status()
                    LOGGER.info(f'File {key} has been stored to Bucket')


def get_xircuits_file():
    """
    :return: the file name and the absolute path for the compiled workflow file
    """
    # check that compiled .xircuits file is correctly passed as argument
    file_arg = sys.argv[1]
    LOGGER.info(f'Identified the executable file: {file_arg}')

    if os.path.exists(file_arg):
        full_path = os.path.abspath(file_arg)
    else:
        LOGGER.error(f"Cannot find the executable file: {file_arg}")
        full_path = None

    filename = os.path.basename(file_arg)

    return filename, full_path


def launch_job(site, project, workflow_file_name, workflow_file_path, files_to_upload, do_stage_out=False,
               filesystem=None, python=None, libraries=None, modules=None):
    """
    Submit a job to a EBRAINS HPC site
    :param site: unicore site
    :param workflow_file_name: base name of compiled workflow file
    :param workflow_file_path: absolute path of compiled workflow file
    :param files_to_upload: list of additional files that need to be sent to the HPC server
    :return: None
    """
    inputs = [workflow_file_path]
    if files_to_upload:
        inputs.extend(files_to_upload)

    submitter = PyunicoreSubmitter(site, project)
    submitter.set_hpc_settings(filesystem, python, libraries, modules)
    submitter.submit_job(workflow_file_name, inputs, do_stage_out)


if __name__ == '__main__':
    workflow_name, workflow_path = get_xircuits_file()
    LOGGER.info("Preparing job...")
    files_to_upload = get_files_to_upload(xircuits_file_path=workflow_path)
    project_arg = sys.argv[3]
    if project_arg == 'NONE':
        LOGGER.error(f"Please provide the HPC project to run this job within, stopping execution.")
    else:
        site_arg = sys.argv[2]
        stage_out_arg = sys.argv[4]
        filesystem_arg = sys.argv[5] if sys.argv[5] != 'NONE' else None
        python_arg = sys.argv[6] if sys.argv[6] != 'NONE' else None
        modules_arg = sys.argv[7] if sys.argv[7] != 'NONE' else None
        libraries_arg = sys.argv[8] if sys.argv[8] != 'NONE' else None
        libraries = libraries_arg.replace(",", " ")
        do_stage_out = True if stage_out_arg == 'on' else False
        launch_job(site=site_arg, project=project_arg, workflow_file_name=workflow_name,
                   workflow_file_path=workflow_path, files_to_upload=files_to_upload, do_stage_out=do_stage_out,
                   filesystem=filesystem_arg, python=python_arg, libraries=libraries, modules=modules_arg)



import os
import posixpath
import json
import toml
from configparser import ConfigParser

def parse_gitmodules(gitmodules_path):
    config = ConfigParser()
    config.read(gitmodules_path)
    modules = []
    for section in config.sections():
        path = config.get(section, 'path', fallback=None)
        url = config.get(section, 'url', fallback=None)
        # Extract the library ID from the path
        library_id = path.replace('xai_components/xai_', '').upper()
        if path and url:
            modules.append({
                'path': posixpath.normpath(path),
                'url': url,
                'library_id': library_id 
            })
    return modules

def parse_toml_file(toml_path):
    try:
        with open(toml_path, 'r') as toml_file:
            data = toml.load(toml_file)
        return data
    except Exception as e:
        print(f"Error parsing TOML file at {toml_path}: {e}")
        return None 

def read_file_lines_to_list(file_path):
    if not os.path.exists(file_path):
        return []
    with open(file_path, 'r') as file:
        return [line.strip() for line in file.readlines()]

def is_installed_component(directory_path):
    # Check for an __init__.py file to determine if the directory is an installed component
    return os.path.isfile(os.path.join(directory_path, '__init__.py'))

def extract_library_info(lib_path, base_path, status="remote"):
    # If __init__.py exists, confirm the component is installed
    if is_installed_component(lib_path):
        status = "installed"

    library_id = posixpath.basename(lib_path).replace('xai_', '').replace('xai-', '').upper()
    lib_info = {
        "name": posixpath.basename(lib_path),
        "library_id": library_id,
        "local_path": posixpath.join(base_path, posixpath.relpath(lib_path, start=base_path)),
        "status": status
    }

    toml_path = os.path.join(lib_path, 'pyproject.toml')
    if os.path.exists(toml_path):
        toml_data = parse_toml_file(toml_path)
        if toml_data:
            lib_info.update({
                "version": toml_data["project"].get("version", "N/A"),
                "description": toml_data["project"].get("description", "No description available."),
                "authors": toml_data["project"].get("authors", []),
                "license": toml_data["project"].get("license", "N/A"),
                "readme": toml_data["project"].get("readme", None),
                "repository": toml_data["project"].get("repository", None),
                "keywords": toml_data["project"].get("keywords", []),
                "requirements": toml_data["project"].get("dependencies", []),
                "default_example_path": toml_data["tool"].get("xircuits", {}).get("default_example_path", None),
            })

    return lib_info

def generate_component_library_config(base_path="xai_components", gitmodules_path=".gitmodules"):

    if not os.path.exists(gitmodules_path):
        # Construct the .xircuits/.gitmodules path
        gitmodules_path = posixpath.join('.xircuits', '.gitmodules')
    
    libraries = {}
    library_id_map = {}  # Map library IDs to library info
    
    # Parse submodules first and set them as "remote"
    if os.path.exists(gitmodules_path):
        submodules = parse_gitmodules(gitmodules_path)
        for submodule in submodules:
            submodule_path = posixpath.normpath(submodule['path'])
            library_info = {
                "name": posixpath.basename(submodule_path),
                "library_id": submodule['library_id'],  # Use the library ID from the submodule info
                "repository": submodule['url'],
                "local_path": submodule_path,
                "status": "remote"
            }
            libraries[submodule_path] = library_info
            library_id_map[submodule['library_id']] = library_info

    def explore_directory(directory, base_path):
        for item in os.listdir(directory):
            full_path = posixpath.normpath(posixpath.join(directory, item))
            if os.path.isdir(full_path) and item.startswith("xai_"):
                lib_info = extract_library_info(full_path, base_path)
                if lib_info:  # If a valid pyproject.toml is found
                    if lib_info['library_id'] in library_id_map:  # Match by library ID
                        # Update the existing entry with the new info
                        library_id_map[lib_info['library_id']].update(lib_info)
                    else:
                        libraries[full_path] = lib_info  # Add new library info
                # Recursively explore subdirectories
                explore_directory(full_path, base_path)

    explore_directory(base_path, base_path)

    return {"libraries": list(libraries.values())}

def save_component_library_config(filename=".xircuits/component_library_config.json"):
    libraries_data = generate_component_library_config()
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w') as json_file:
        json.dump(libraries_data, json_file, indent=4)
    return libraries_data

def get_component_library_config(filename=".xircuits/component_library_config.json"):
    if os.path.exists(filename):
        try:
            with open(filename, 'r') as json_file:
                return json.load(json_file)
        except Exception as e:
            print(f"Error reading JSON file at {filename}: {e}")
            return None
    else:
        return save_component_library_config(filename)

if __name__ == "__main__":
    save_component_library_config()



from argparse import Namespace
from typing import TypeVar, Generic, Tuple, NamedTuple, Callable, List
from copy import deepcopy

from xai_components.logger.builder import get_logger

LOGGER = get_logger(__name__)

T = TypeVar('T')

class OutArg(Generic[T]):
    def __init__(self, value: T = None, getter: Callable[[T], any] = lambda x: x) -> None:
        self._value = value
        self._getter = getter

    @property
    def value(self):
        return self._getter(self._value)

    @value.setter
    def value(self, value: T):
        self._value = value

    def connect(self, ref: 'OutArg[T]'):
        self._value = ref
        self._getter = lambda x: x.value

    def __copy__(self):
        return type(self)(self._value, self._getter)

    def __deepcopy__(self, memo):
        id_self = id(self)
        _copy = memo.get(id_self)
        if _copy is None:
            _copy = type(self)(
                deepcopy(self._value, memo),
                deepcopy(self._getter, memo)
            )
            memo[id_self] = _copy
        return _copy


class InArg(Generic[T]):
    def __init__(self, value: T = None, getter: Callable[[T], any] = lambda x: x) -> None:
        self._value = value
        self._getter = getter

    @property
    def value(self):
        return self._getter(self._value)

    @value.setter
    def value(self, value: T):
        self._value = value

    def connect(self, ref: OutArg[T]):
        self._value = ref
        self._getter = lambda x: x.value

    def __copy__(self):
        return type(self)(self._value, self._getter)

    def __deepcopy__(self, memo):
        id_self = id(self)
        _copy = memo.get(id_self)
        if _copy is None:
            _copy = type(self)(
                deepcopy(self._value, memo),
                deepcopy(self._getter, memo)
            )
            memo[id_self] = _copy
        return _copy

class InCompArg(Generic[T]):
    def __init__(self, value: T = None, getter: Callable[[T], any] = lambda x: x) -> None:
        self._value = value
        self._getter = getter

    @property
    def value(self):
        return self._getter(self._value)

    @value.setter
    def value(self, value: T):
        self._value = value

    def connect(self, ref: OutArg[T]):
        self._value = ref
        self._getter = lambda x: x.value

    def __copy__(self):
        return type(self)(self._value, self._getter)

    def __deepcopy__(self, memo):
        id_self = id(self)
        _copy = memo.get(id_self)
        if _copy is None:
            _copy = type(self)(
                deepcopy(self._value, memo),
                deepcopy(self._getter, memo)
            )
            memo[id_self] = _copy
        return _copy

def xai_component(*args, **kwargs):
    # Passthrough element without any changes.
    # This is used for parser metadata only.
    if len(args) == 1 and callable(args[0]):
        # @xai_components form
        return args[0]
    else:
        # @xai_components(...) form
        def passthrough(f):
            return f
        return passthrough

class ExecutionContext:
    args: Namespace

    def __init__(self, args: Namespace):
        self.args = args

class BaseComponent:
    def __init__(self):
        all_ports = self.__annotations__
        for key, type_arg in all_ports.items():
            if hasattr(type_arg, '__origin__'):
                port_class = type_arg.__origin__
                port_type = type_arg.__args__[0]
                if port_class in (InArg, InCompArg, OutArg):
                    if hasattr(port_type, 'initial_value'):
                        port_value = port_type.initial_value()
                    else:
                        port_value = None

                    if hasattr(port_type, 'getter'):
                        port_getter = port_type.getter
                    else:
                        port_getter = lambda x: x
                    setattr(self, key, port_class(port_value, port_getter))
                else:
                    setattr(self, key, None)
            else:
                setattr(self, key, None)

    @classmethod
    def set_execution_context(cls, context: ExecutionContext) -> None:
        cls.execution_context = context

    def execute(self, ctx) -> None:
        pass

    def do(self, ctx) -> 'BaseComponent':
        pass

    def __copy__(self):
        _copy = type(self)()
        for key, type_arg in self.__dict__.items():
            setattr(_copy, key, getattr(self, key))
        return _copy

    def __deepcopy__(self, memo):
        id_self = id(self)
        _copy = memo.get(id_self)
        if _copy is None:
            _copy = type(self)()
            memo[id_self] = _copy
            for key, type_arg in self.__dict__.items():
                setattr(_copy, key, deepcopy(getattr(self, key), memo))
        return _copy

class Component(BaseComponent):
    next: BaseComponent

    def do(self, ctx) -> BaseComponent:
        LOGGER.info(f"\nExecuting: {self.__class__.__name__}")
        self.execute(ctx)

        return self.next

    def debug_repr(self) -> str:
        return "<h1>Component</h1>"


class SubGraphExecutor:
    
    def __init__(self, component):
        self.comp = component
        
    def do(self, ctx):
        comp = self.comp
        
        while comp is not None:
            comp = comp.do(ctx)
        return None


def execute_graph(args: Namespace, start: BaseComponent, ctx) -> None:
    BaseComponent.set_execution_context(ExecutionContext(args))

    if 'debug' in args and args['debug']:
        import pdb
        pdb.set_trace()

        current_component = start
        next_component = current_component.do(ctx)
        while next_component:
            current_component = next_component
            next_component = current_component.do(ctx)
    else:
        next_component = start.do(ctx)
        while next_component:
            next_component = next_component.do(ctx)


class secret:
    pass

class message(NamedTuple):
    role: str
    content: str

class chat(NamedTuple):
    messages: List[message]

class dynalist(list):
    def __init__(self, *args):
        super().__init__(args)

    @staticmethod
    def getter(x):
        if x is None:
            return []
        return [item.value if isinstance(item, (InArg, OutArg)) else item for item in x]

class dynatuple(tuple):
    def __init__(self, *args):
        super().__init__(args)
    @staticmethod
    def getter(x):
        if x is None:
            return tuple()
        def resolve(item):
            if isinstance(item, (InArg, InCompArg,OutArg)):
                return item.value
            else:
                return item
        return tuple(resolve(item) for item in x)

def parse_bool(value):
    if value is None:
        return None
    if value.lower() in ('true', 't', 'yes', 'y', '1'):
        return True
    elif value.lower() in ('false', 'f', 'no', 'n', '0'):
        return False


import os
import sys
import json
import subprocess
from pathlib import Path

def get_installed_packages():
    """Return a set of installed packages."""
    result = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
    packages = {package.split('==')[0] for package in result.decode().splitlines()}
    return packages

def check_requirements_installed(installed_packages, requirements):
    """Check if all required packages are installed."""
    required_packages = {pkg.strip() for pkg in requirements}
    missing_packages = required_packages - installed_packages
    return len(missing_packages) == 0, missing_packages

def list_component_library():
    config_path = Path(".xircuits/component_library_config.json")
    if not config_path.exists():
        print("Component library config file not found.")
        return

    with open(config_path, 'r') as file:
        config = json.load(file)
    libraries = config.get("libraries", [])

    print("Checking installed packages... This might take a moment.")
    installed_packages = get_installed_packages()

    fully_installed = []
    incomplete_installation = []

    for lib in libraries:
        lib_name = lib["name"]
        status = lib["status"]
        requirements = lib.get("requirements", [])
        
        if status != "remote":
            is_fully_installed, missing_packages = check_requirements_installed(installed_packages, requirements)
            if is_fully_installed:
                fully_installed.append(lib_name)
            else:
                # Append a star to libraries with incomplete installations
                incomplete_installation.append(f"{lib_name} [*]")

    total_installed = len(fully_installed) + len(incomplete_installation)
    print(f"\nInstalled component libraries({total_installed}):")
    for lib in sorted(fully_installed):
        print(f" - {lib}")
    for lib in sorted(incomplete_installation):
        print(f" - {lib}")

    if incomplete_installation:
        print("\n[*] indicates an incomplete installation.")

    # Handle remote libraries separately
    remote = [lib["name"] for lib in libraries if lib["status"] == "remote"]
    if remote:
        print(f"\nRemote component libraries({len(remote)}):")
        for lib in sorted(remote):
            print(f" - {lib}")
        print("\nYou can install remote libraries using 'xircuits install <libName>'.\n\n")

if __name__ == "__main__":
    list_component_library()



from .list_library import list_component_library
from .install_fetch_library import install_library, fetch_library, get_library_config, build_library_file_path_from_config
from .generate_component_library_config import save_component_library_config, get_component_library_config
from .create_library import create_or_update_library